from . import messages, users
from .utils import load_peer


def download(api):
    """
    Загружает базовую информацию о всех переписках пользователя

    Args:
        api (vk.API): Объект, через который происходит обращение к
            методам VK API
    """
    # Получаем часть переписок
    res = api.messages.getConversations(count=200)
    peers = [item['conversation'] for item in res['items']]

    processed = len(peers)

    # Повторяем действия выше, пока все переписки не будут загружены
    while processed < res['count']:
        res = api.messages.getConversations(offset=processed, count=200)
        peers += [item['conversation'] for item in res['items']]

    return peers


class Peer:
    """
    Класс для представления всей переписки пользователя из JSON

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        peer_id (int): Идентификатор переписки, которую нужно представить
    """

    def __init__(self, out_dir, peer_id):
        # Загружаем всю информацию из JSON
        peer = load_peer(out_dir, peer_id)

        usernames = users.parse(peer)

        # Сохраняем информацию о переписке
        self.info = peer['info']

        # Парсим все сообщения переписки
        self.msgs = messages.parse(peer, usernames)

        # Сохраняем название переписки
        if self.info['peer']['type'] == 'chat':
            self.title = self.info['chat_settings']['title']
        else:
            self.title = usernames[self.info['peer']['id']]



from collections import deque


def download(api, peer):
    """
    Загружает имена участников переписки

    Args:
        api (vk.API): Объект, через который происходит обращение к
            методам VK API
        peer (dict): Объект (словарь) переписки, в который необходимо
            сохранить скачанные имена
    """
    # Идентификаторы пользователей и групп
    user_ids = set()
    group_ids = set()

    # Обход сообщений в ширину
    queue = deque(peer['messages'])

    if peer['info']['peer']['id'] < 2000000000:
        queue.append({'from_id': peer['info']['peer']['id']})

    while queue:
        msg = queue.popleft()

        # Добавляем идентификатор отправителя
        from_ids = [msg['from_id']]

        # Добавляем идентификатор пользователя, над которым совершается
        # действие, если у сервисного сообщения такой есть
        if 'action' in msg and 'member_id' in msg['action']:
            from_ids.append(msg['action']['member_id'])

        for from_id in from_ids:
            # Идентификаторы групп передаются как положительные числа
            eff_id = str(abs(from_id))

            # Если идентификатор < 0, добавляем к группам
            if from_id < 0:
                group_ids.add(eff_id)

            # Если идентификатор > 0, добавляем к пользователям
            else:
                user_ids.add(eff_id)

        # Если у сообщения есть пересланные, то добавляем их в очередь
        if 'fwd_messages' in msg:
            queue.extend(msg['fwd_messages'])

        if 'reply_message' in msg:
            queue.append(msg['reply_message'])

    # Получаем имена пользователей и групп
    res = {'users': [], 'groups': []}

    if user_ids:
        res['users'].extend(api.users.get(user_ids=','.join(user_ids)))

    if group_ids:
        res['groups'].extend(api.groups.getById(group_ids=','.join(group_ids)))

    # Сохраняем имена
    peer['members'] = res


def parse(peer):
    """
    Получает имена пользователей из JSON для дальнейшей работы

    Args:
        peer (dict): Объект (словарь) переписки, в который уже
            были сохранены скачанные сообщения

    Returns:
        Dict[int, str]: Словарь для получения имени пользователя по его ID
    """
    users = {}

    # Обрабатываем пользователей
    for json in peer['members']['users']:
        users[json['id']] = json['first_name']

        # Если пользователь был удален, то у него не будет фамилии
        if json.get('deactivated') != 'deleted':
            users[json['id']] += ' ' + json['last_name']

    # Обрабатываем группы
    for json in peer['members']['groups']:
        users[-json['id']] = json['name']

    return users



from argparse import ArgumentParser
from os import getenv

from .utils import parse_peer_ids


def parse_args():
    """
    Создает объект парсера и парсит аргументы

    Returns:
        argparse.Namespace: Набор аргументов парсера
    """
    parser = ArgumentParser()

    parser.add_argument(
        '-o',
        dest='out_dir',
        metavar='OUTDIR',
        default='vkms-result',
        help='Output directory where the materials are saved'
    )
    parser.add_argument(
        '-i',
        '--include',
        type=parse_peer_ids,
        default=set(),
        help='Comma-separated list of peer IDs to process. '
             'If not specified, all peers will be processed'
    )
    parser.add_argument(
        '-e',
        '--exclude',
        type=parse_peer_ids,
        default=set(),
        help='Comma-separated list of peer IDs that DON\'T need to be processed. '
             'Ignored if the --include parameter is specified'
    )

    subparsers = parser.add_subparsers(
        dest='action',
        required=True,
        help='One of the actions to be performed'
    )

    parser_dump = subparsers.add_parser(
        'dump',
        help='Save only the necessary information (VK API method outputs) '
             'in a machine-friendly format (JSON) for further processing'
    )

    parser_dump.add_argument(
        '--token',
        help='Access token of an account with scope of messages. '
             'You also can pass it as an env variable ACCESS_TOKEN'
    )
    parser_dump.add_argument(
        '-t',
        '--threads',
        type=int,
        default=2,
        help='Number of threads to download peers, defaults to 2'
    )
    parser_dump.add_argument(
        '-m',
        '--max-msgs',
        metavar='MAX',
        type=int,
        default=75000,
        help='Maximum number of messages to be saved in each conversation. '
             'Increasing the parameter will consume more RAM. '
             'Defaults to 75000'
    )

    parser_parse = subparsers.add_parser(
        'parse',
        help='Convert machine-friendly JSON format to human-readable'
    )

    parser_parse.add_argument(
        '-f',
        dest='fmt',
        choices=('txt',),
        default='txt',
        metavar='FORMAT',
        help='An easy-to-read format in which received messages must be '
             'converted. Supported formats: %(choices)s'
    )

    parser_atch = subparsers.add_parser(
        'atch',
        help='Download messages attachments (photos, audio, etc.)'
    )

    parser_atch.add_argument(
        '-t',
        '--threads',
        type=int,
        default=8,
        help='Number of threads to download peers, defaults to 8'
    )

    args = parser.parse_args()

    if args.action == 'dump':
        token = getenv('ACCESS_TOKEN') or args.token

        if token is None:
            parser.error(
                'You must define token as an ACCESS_TOKEN env variable '
                'or using the --token argument'
            )

        args.token = token

    return args



import logging
from os import makedirs

from . import actions
from .argparser import parse_args


def _setup_logging(out_dir):
    logging.getLogger('vk').setLevel(logging.WARNING)
    logging.basicConfig(
        filename=f'{out_dir}/logs.txt',
        filemode='a',
        format='[%(levelname)s] (%(asctime)s.%(msecs)03d - '
               '%(module)s - %(threadName)s): %(message)s',
        level=logging.INFO,
        datefmt='%m/%d/%Y %H:%M:%S'
    )


def main():
    """
    Входная точка программы
    """
    args = parse_args()

    makedirs(f'{args.out_dir}/.json/', exist_ok=True)

    _setup_logging(args.out_dir)
    logging.info('VKMS started')
    logging.info(f'Arguments: {args}')

    if args.action == 'dump':
        actions.dump(
            args.out_dir,
            args.include,
            args.exclude,
            args.token,
            args.threads,
            args.max_msgs
        )

    elif args.action == 'parse':
        actions.parse(args.out_dir, args.include, args.exclude, args.fmt)

    elif args.action == 'atch':
        actions.atch(args.out_dir, args.include, args.exclude, args.threads)

    logging.info('VKMS completed\n')



from json import dump, load

# Названия месяцев на русском
months = [
    ' ', 'января', 'февраля', 'марта',
    'апреля', 'мая', 'июня', 'июля',
    'августа', 'сентября', 'октября',
    'ноября', 'декабря'
]


def load_peer(out_dir, peer_id):
    """
    Загружает переписку из скачанного JSON по указанному идентификатору

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        peer_id (int): Идентификатор переписки, JSON которой
            необходимо загрузить
    """
    with open(f'{out_dir}/.json/{peer_id}.json', 'r') as file:
        return load(file)


def dump_peer(out_dir, peer):
    """
    Выгружает объект переписки в JSON

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        peer (dict): Объект (словарь) переписки, которую необходимо
            выгрузить в JSON
    """
    with open(f"{out_dir}/.json/{peer['info']['peer']['id']}.json", 'w') as file:
        dump(peer, file)


def parse_peer_ids(peer_ids):
    """
    Парсит разделенный запятыми список идентификаторов переписок в
    множество, обрабатывая особые случаи

    Args:
        peer_ids (str): Разделенный запятыми список идентификаторов переписок

    Returns:
        Set[int]: Множество идентификаторов переписок
    """
    res = set()

    for peer_id in peer_ids.split(','):
        if not peer_id:
            continue

        # Сокращенный вариант идентификатора беседы
        if peer_id[0] == 'c' and peer_id[1:].isdigit():
            res.add(2000000000 + int(peer_id[1:]))

        # Идентификатор переписки с пользователем и группой
        elif peer_id.isdigit() or peer_id[0] == '-' and peer_id[1:].isdigit():
            res.add(int(peer_id))

        else:
            raise ValueError('Invalid peer ID: ' + peer_id)

    return res



import logging
from os import listdir
from threading import Thread
from time import sleep

import vk
from vk.exceptions import VkAPIError

from . import attachments, messages, peers, saver, users
from .utils import dump_peer


def dump(out_dir, include, exclude, token, nthreads, max_msgs):
    """
    Скачивает указанные переписки в формате JSON (результаты обращений к VK API)

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        include (set): Множество идентификаторов переписок, которые нужно сохранить
        exclude (set): Множество идентификаторов переписок, которые не нужно сохранять
        token (str): Токен доступа к VK API
        nthreads (int): Количество потоков, загружающих переписки
    """
    # Получаем объект для работы с VK API
    class API(vk.API):
        def on_api_error_6(self, request):
            sleep(1)
            return self.send(request)

        def get_captcha_key(self, request):
            return input(f'Captcha needed ({request.api_error.captcha_img}): ')

    api = API(access_token=token, v='5.131')

    # Загружаем информацию о владельце страницы
    try:
        account = api.users.get()[0]
    except VkAPIError:
        print('Failed to load base information. '
              f'See logs for details: {out_dir}/logs.txt')

        return

    # Загружаем информацию о всех переписках пользователя
    peers_info = peers.download(api)

    # Создаем словарь вида id-переписка для быстрого доступа к JSON'у конкретной переписки
    peer_by_id = {peer['peer']['id']: peer for peer in peers_info}

    # Создаем множество идентификаторов для последующей выборки нужных
    peer_ids = set(peer_by_id.keys())

    # Выбираем нужные переписки
    if include:
        peer_ids &= include
    elif exclude:
        peer_ids -= exclude

    peer_ids = list(peer_ids)
    peer_ids_len = len(peer_ids)

    logging.info(f"Peers: {', '.join(map(str, peer_ids))}")

    def dump_thread():
        """Поток для загрузки переписки"""

        # Работает, пока остались необработанные переписки
        while True:
            try:
                peer_id = peer_ids.pop()
            except IndexError:
                return

            logging.info(f'Processing peer {peer_id}')

            # Сохраняем информацию о переписке и владельце страницы
            peer = {'info': peer_by_id[peer_id]}
            peer['info']['account'] = account

            try:
                # Сохраняем все сообщения и информацию об участниках переписки
                messages.download(api, peer_id, peer, max_msgs)
                users.download(api, peer)

            except VkAPIError as e:
                logging.error(f'Downloading peer {peer_id} failed: {e}')
                return

            # Записываем все в JSON
            dump_peer(out_dir, peer)

    # Список всех потоков
    tds = []

    # Создаем N потоков
    for _ in range(min(nthreads, peer_ids_len)):
        td = Thread(target=dump_thread, daemon=True)
        td.start()
        tds.append(td)

    # Ждем, пока все переписки будут скачаны
    while sum(td.is_alive() for td in tds):
        print(f'{round((peer_ids_len - len(peer_ids)) / peer_ids_len * 100)}%', end='\r')

    # Сливаем потоки
    for td in tds:
        td.join()

    print('100%')


def parse(out_dir, include, exclude, fmt):
    """
    Сохраняет полученные переписки в удобном для чтения формате

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        include (set): Множество идентификаторов переписок, которые нужно сохранить
        exclude (set): Множество идентификаторов переписок, которые не нужно сохранять
        fmt (str): Формат, в котором следует сохранять переписки
    """
    # Получаем идентификаторы всех скачанных переписок
    peer_ids = {int(file.rstrip('.json')) for file in listdir(f'{out_dir}/.json/')}

    # Выбираем нужные переписки
    if include:
        peer_ids &= include
    elif exclude:
        peer_ids -= exclude

    processed = 0

    logging.info(f"Peers: {', '.join(map(str, peer_ids))}")

    # Обрабатываем каждую переписку отдельно
    for peer_id in peer_ids:
        logging.info(f'Processing peer {peer_id}')

        print(f'{round(processed / len(peer_ids) * 100)}%', end='\r')

        try:
            peer = peers.Peer(out_dir, peer_id)
        except Exception:
            logging.error(f'Loading peer {peer_id} failed: JSON may be corrupted')
        else:
            saver.save(out_dir, fmt, peer)

        processed += 1

    print('100%')


def atch(out_dir, include, exclude, nthreads):
    """
    Скачивает вложения указанных переписок

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        include (set): Множество идентификаторов переписок, которые нужно обработать
        exclude (set): Множество идентификаторов переписок, которые не нужно обрабатывать
        nthreads (int): Количество потоков, загружающих вложения
    """
    # Получаем идентификаторы всех скачанных переписок
    peer_ids = {int(file.rstrip('.json')) for file in listdir(f'{out_dir}/.json/')}

    # Выбираем нужные переписки
    if include:
        peer_ids &= include
    elif exclude:
        peer_ids -= exclude

    logging.info(f"Peers: {', '.join(map(str, peer_ids))}")

    peer_ids_cnt = 0

    # Обрабатываем каждую переписку отдельно
    for peer_id in peer_ids:
        logging.info(f'Processing peer {peer_id}')

        print(f'{round(peer_ids_cnt / len(peer_ids) * 100)}%', end='\r')

        try:
            peer = peers.Peer(out_dir, peer_id)
        except Exception:
            logging.error(f'Loading peer {peer_id} failed: JSON may be corrupted')
        else:
            attachments.download(out_dir, peer, nthreads)

        peer_ids_cnt += 1

    print('100%')



import datetime
from typing import Dict, List, Tuple

from .attachments import Attachment, gen_attachment
from .utils import months


def download(api, peer_id, peer, max_msgs):
    """
    Загружает сообщения переписки. В начале идут старые сообщения, в конце - новые

    Args:
        api (vk.API): Объект, через который происходит обращение к
            методам VK API
        peer_id (int): Идентификатор переписки, сообщения которой
            необходимо скачать
        peer (dict): Объект (словарь) переписки, в который необходимо
            сохранить скачанные сообщения
    """
    # Получаем часть сообщений
    res = api.messages.getHistory(count=200, peer_id=peer_id)
    msgs = res['items']

    processed = len(msgs)

    # Повторяем действия выше, пока все сообщения не будут загружены
    while processed < min(res['count'], max_msgs):
        res = api.messages.getHistory(offset=processed, count=200, peer_id=peer_id)
        msgs += res['items']

        processed += len(res['items'])

    # Переворачиваем сообщения, чтобы в начале
    # оказались старые, а в конце - новые
    msgs.reverse()

    # Сохраняем сообщения
    peer['messages'] = msgs


# Шаблоны текстов сервисных действий в формате тип-текст
#
# Args:
#     member: Имя инициатора действия
#     user: Имя пользователя, над которым выполняется действие
#     text: Новое название беседы
#     msg: Текст закрепленного сообщения
#
_actions: Dict[str, str] = {
    'chat_photo_update': '{member} обновил(-a) фотографию беседы',
    'chat_photo_remove': '{member} удалил(-a) фотографию беседы',
    'chat_create': '{member} создал(-а) беседу "{text}"',
    'chat_title_update': '{member} изменил(-а) название беседы на "{text}"',
    'chat_invite_user': '{member} пригласил(-а) {user}',
    'chat_kick_user': '{member} исключил(-а) {user}',
    'chat_pin_message': '{member} закрепил(-а) сообщение "{msg}"',
    'chat_unpin_message': '{member} открепил(-а) сообщение',
    'chat_invite_user_by_link': '{member} присоединился(-ась) к беседе по ссылке',
    'conversation_style_update': '{member} изменил(-а) оформление чата',
    'chat_screenshot': '{member} сделал(-а) скриншот беседы',
    '_chat_leave_user': '{member} вышел(-а) из беседы'
}


class Message:
    """
    Класс для представления объекта `message` из JSON
    Документация: [https://dev.vk.com/reference/objects/message]

    Args:
        json (dict): Объект сообщения полученный ранее благодаря VK API
        usernames (dict): Словарь для получения имени пользователя по его ID
    """

    def __init__(self, json, usernames):
        # Идентификатор сообщения. Вычисляется немного необычным путем, т.к.
        # сообщения полученные из объектов `reply_message` и `fwd_messages`
        # могут иметь другие поля ID
        self.id: tuple = Message.get_id_by_json(json)

        # Получение имени отправителя по его ID
        self.username: str = usernames[json['from_id']]

        # Получение времени отправления сообщения из формата Unixtime
        self.date = datetime.datetime.fromtimestamp(json['date'])

        # Информация о сервисном действии, `self.action` будет содержать
        # уже готовый текст или `None`
        self.action: str | None = None

        if 'action' in json:
            # Если пользователь исключил сам себя, значит он вышел из беседы
            if (json['action']['type'] == 'chat_kick_user'
                    and json['from_id'] == json['action']['member_id']):
                json['action']['type'] = '_chat_leave_user'

            # Генерация текста из шаблона
            self.action = _actions.get(json['action']['type'], 'unknown action').format(
                user=usernames.get(json['action'].get('member_id')),
                member=self.username,
                text=json['action'].get('text'),
                msg=json['action'].get('message')
            )

        # Текст сообщения (может быть пустым)
        self.text: str = json['text']

        # Объект сообщения, в ответ на которое было отправлено текущее
        self.reply_msg: Message = None

        if 'reply_message' in json:
            self.reply_msg = gen_message(json['reply_message'], usernames)

        # Флаг, является ли актуальным сообщение (для исчезающих сообщений)
        self.is_expired = json.get('is_expired')

        # Флаг, было ли сообщение отредактировано
        self.is_edited = 'update_time' in json

        # Координаты геолокации (если есть)
        self.geo = None

        if 'geo' in json:
            self.geo = json['geo']['coordinates'].values()

        # Список пересланных сообщений, вместе образуют дерево
        self.fwd_msgs: List[Message] = [
            gen_message(fwd_msg_json, usernames)
            for fwd_msg_json in json.get('fwd_messages', [])
        ]

        # Список медиавложений сообщения (фото, аудио и т.п.)
        self.atchs: List[Attachment] = [
            gen_attachment(atch_json)
            for atch_json in json['attachments']
        ]

    @staticmethod
    def get_id_by_json(json):
        """
        Функция для получения идентификатора сообщения из JSON

        Args:
            json (dict): Объект сообщения полученный ранее благодаря VK API

        Returns:
            Tuple[int, int, int]: Идентификатор
        """
        return (json['date'], json['from_id'], json['conversation_message_id'])

    def full_date(self):
        """
        Возвращает полную дату (день, месяц, год) отправки сообщения

        Returns:
            str: Полная дата
        """
        return self.date.strftime('%d {} %Y'.format(months[self.date.month]))

    def time(self):
        """
        Возвращает время (час, минута) отправки сообщения

        Returns:
            str: Время отправки
        """
        return self.date.strftime('%H:%M')


# Словарь для кеширования сообщений в формате id-сообщение
_msgs: Dict[Tuple[int, int, int], Message] = {}


def parse(peer, usernames):
    """
    Парсит сообщения из JSON в объекты `Message` для дальнейшей работы

    Args:
        peer (dict): Объект (словарь) переписки, в который уже
            были сохранены скачанные сообщения
        usernames (dict): Словарь для получения имени пользователя по его ID

    Returns:
        List[Message]: Список объектов сообщений
    """
    # Очистка кеша сообщений
    _msgs.clear()

    # Генерация сообщений
    return [gen_message(msg, usernames) for msg in peer['messages']]


def gen_message(json, usernames):
    """
    Генерирует и кеширует созданное сообщение по его ID

    Args:
        json (dict): Объект сообщения полученный ранее благодаря VK API
        usernames (dict): Словарь для получения имени пользователя по его ID

    Returns:
        Message: Сгенерированный объект сообщения
    """
    msg_id = Message.get_id_by_json(json)

    if msg_id not in _msgs:
        _msgs[msg_id] = Message(json, usernames)

    return _msgs[msg_id]



from os import makedirs
from re import sub

from pathvalidate import sanitize_filename


def save(out_dir, fmt, peer):
    """
    Конвертирует и сохраняет сообщения в удобном для чтения формате

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        fmt (str): Формат, в котором нужно сохранить переписку
        peer (peers.Peer): Объект переписки
    """
    globals()['save_' + fmt](out_dir, peer)


def save_txt(out_dir, peer):
    """
    Сохраняет переписку в формате TXT. Идея взята у hikiko4ern
    Ссылка: [https://github.com/hikiko4ern/vk_dump]
    """
    # Создаем папку для хранения переписки
    makedirs(f'{out_dir}/dialogs/txt/', exist_ok=True)

    path = '{out_dir}/dialogs/txt/{title}_{peer_id}.txt'.format(
        out_dir=out_dir,
        title=sanitize_filename(peer.title, replacement_text='_'),
        peer_id=peer.info['peer']['id']
    )

    # Сохраняем конвертированный текст
    with open(path, 'w') as file:
        file.write(convert_txt(peer.msgs, peer.info['account']['id']))


def convert_txt(msgs, account_id, is_reply=False):
    # Буффер текста переписки
    buf = ''

    prev_msg = None

    for msg in msgs:
        # Если предыдущего сообщения не было, пишем дату текущего сообщения
        if prev_msg is None:
            buf += f'        [{msg.full_date()}]\n'

        # Если предыдущее и текущее сообщения были отправлены в разные дни,
        # пишем дату текущего сообщения
        elif msg.date.day - prev_msg.date.day >= 1:
            buf += f'\n        [{msg.full_date()}]\n'
            prev_msg = None

        # Буффер текста текущего сообщения (каждый элемент - новая строка)
        text = []

        # Если сообщение является сервисным, добавляем текст действия
        if msg.action:
            text.append(f'[{msg.action}]')

        # Если сообщение отправлено в ответ на другое, обрабатываем другое
        # сообщение отдельно
        if not is_reply and msg.reply_msg:
            convert_txt([msg.reply_msg], account_id, True)

            for line in msg.reply_msg.text_parsed:
                text.append('{username} {line}'.format(
                    username=(
                        msg.reply_msg.username + '>'
                        if line == msg.reply_msg.text_parsed[0] else
                        ' ' * (len(msg.reply_msg.username) + 1)
                    ),
                    line=line
                ))

        # Добавляем текст самого сообщения (если есть), обрабатывая обращения
        if msg.text:
            text.extend([
                sub(r'\[(?:club|id)\d+\|([^\]]+)\]', r'\1', line)
                for line in msg.text.split('\n')
            ])

        # Если есть пересланные сообщения, обрабатываем их рекурсивно
        for line in filter(None, convert_txt(msg.fwd_msgs, account_id).split('\n')):
            text.append('| ' + line)

        # Добавляем место/геолокацию (если есть)
        if msg.geo:
            text.append('[место: {}, {}]'.format(*msg.geo))

        # Обрабатываем медиавложения
        for atch in msg.atchs:
            if atch.tp == 'photo':
                text.append(f'[фото: {atch.filename}]')
            elif atch.tp == 'video':
                text.append(f'[видео: {atch.filename}]')
            elif atch.tp == 'audio':
                text.append(f'[аудио: {atch.title}]')
            elif atch.tp == 'doc':
                text.append(f'[документ: {atch.filename}]')
            elif atch.tp == 'link':
                text.append(f'[ссылка: {atch.title} ({atch.url})]')
            elif atch.tp == 'wall':
                text.append(f'[пост: {atch.url}]')
            elif atch.tp == 'wall_reply':
                text.append(f'[комментарий: {atch.url}]')
            elif atch.tp == 'sticker':
                text.append(f'[стикер: {atch.filename}]')
            elif atch.tp == 'gift':
                text.append(f'[подарок: {atch.filename}]')
            elif atch.tp == 'audio_message':
                text.append(f'[голосвое сообщение: {atch.filename}]')
            elif atch.tp == 'graffiti':
                text.append(f'[граффити: {atch.filename}]')

            elif atch.tp == 'call':
                text.append('[{tp}звонок: {state}{duration}]'.format(
                    tp=atch.call_type,
                    state=atch.get_state(account_id),
                    duration=f' ({atch.duration})' if atch.state == 'reached' else ''
                ))

            elif atch.tp == 'poll':
                pool_text = [f'[опрос: "{atch.title}", {atch.anon}, {atch.mult}]']
                max_len = len(pool_text[0])

                for ans in atch.answers:
                    pool_text.append(
                        '  ["{text}":   {{indent}}{is_voted}{rate}% ({votes}/{all_votes})]'.format(
                            text=ans['text'],
                            is_voted='✓ ' if ans['id'] in atch.answer_ids else '',
                            rate=round(ans['rate']),
                            votes=ans['votes'],
                            all_votes=atch.all_votes
                        )
                    )
                    max_len = max(max_len, len(pool_text[-1]) - len('{indent}'))

                for line in pool_text:
                    text.append(line.format(
                        indent=' ' * (max_len - len(line) + len('{indent}'))
                    ))
            else:
                text.append('[uknown attachment]')

        # Если сообщение является исчезающим
        if msg.is_expired:
            text.append('[Сообщение исчезло]')

        # Если сообщение было отредактировано, добавляем заметку
        elif msg.is_edited:
            text.append('(ред.)')

        # Кэшируем полученный текст
        msg.text_parsed = text

        # Записываем полученный текст в буффер переписки
        for line in text:
            if line == text[0]:
                if prev_msg and prev_msg.username == msg.username:
                    username = ' ' * (len(msg.username) + 1)
                else:
                    username = msg.username + ':'

                buf += f"[{msg.time()}] {username} {line}\n"
            else:
                buf += f"{' ' * 7} {' ' * len(msg.username)}  {line}\n"

        # Обновляем предыдущее сообщение
        prev_msg = msg

    # Возвращаем буффер
    return buf



import logging
import os
from operator import itemgetter
from threading import Thread

from pathvalidate import sanitize_filename
from requests import get
from requests.exceptions import RequestException


def download(out_dir, peer, nthreads):
    """
    Скачивает указанные переписки в формате JSON (результаты обращений к VK API)

    Args:
        out_dir (str): Абсолютный путь к каталогу, в котором находится
            результат работы программы
        peer (peers.Peer): Объект переписки
        nthreads (int): Количество потоков, загружающих вложения
    """
    # Создаем папку для хранения вложений
    os.makedirs(f'{out_dir}/attachments/', exist_ok=True)

    # Создаем список, состоящий только из скачиваемых вложений
    atchs = []

    for msg in peer.msgs:
        for atch in msg.atchs:
            if isinstance(atch, FileAttachment):
                atchs.append(atch)

    def atch_thread():
        """Поток для загрузки вложений"""

        # Работает, пока остались нескачанные вложения
        while True:
            try:
                atch = atchs.pop()
            except IndexError:
                return

            logging.info(f'Downloading attachment {atch.tp}, {atch.filename}')

            # Загружаем вложение
            atch.download(out_dir)

    # Список всех потоков
    tds = []

    # Создаем N потоков
    for _ in range(min(nthreads, len(atchs))):
        td = Thread(target=atch_thread, daemon=True)
        td.start()
        tds.append(td)

    # Ждем, пока все вложения будут скачаны
    for td in tds:
        td.join()


class Attachment:
    """
    Абстрактный класс для представления любого объекта `attachment` из JSON
    Документация: [https://dev.vk.com/reference/objects/attachments-message]

    Args:
        json (dict): Объект вложения, полученный ранее благодаря VK API
    """

    # Уникальный тип вложения
    tp = 'unknown'

    def __init__(self, json):  # noqa: U100
        pass


class FileAttachment(Attachment):
    """
    Абстрактный подкласс `Attachment` для представления вложения,
    которое можно скачать (фото, документ и т.п.)
    """

    # Ссылка на загрузку файла
    url = None

    # Название директории, в которую будут сохраняться вложения этого типа
    pf_dir = None

    # Имя файла, под которым вложение будет сохранено (уникальное для каждого вложения)
    filename = None

    def download(self, out_dir):
        """
        Загружает и сохраняет вложение

        Args:
            out_dir (str): Абсолютный путь к каталогу, в котором находится
                результат работы программы
        """
        # Создаем папку для хранения вложений этого типа
        os.makedirs(f'{out_dir}/attachments/{self.pf_dir}/', exist_ok=True)

        path = f'{out_dir}/attachments/{self.pf_dir}/{self.filename}'

        # Если файл уже скачан, пропускаем его
        if os.path.exists(path):
            return

        attempts = 0

        # Есть 3 попытки для установления соединения
        # (если ошибка произошла на стороне сервера)
        while attempts < 3:
            try:
                with get(self.url, stream=True) as r:
                    if r.status_code >= 500:
                        attempts += 1
                        continue

                    # Если ошибка на нашей стороне - пропускаем это вложение
                    elif r.status_code >= 400:
                        logging.error(
                            'Downloading attachment failed: HTTP code %s. Url: %s',
                            r.status_code, self.url
                        )
                        return

                    # Скачивание происходит порциями (чанками), т.к. максимальный
                    # размер вложения VK - 2ГБ
                    with open(path, 'wb') as file:
                        for chunk in r.iter_content(chunk_size=(1 << 20) * 10):
                            file.write(chunk)

                        return

            except RequestException:
                attempts += 1

        logging.error(
            'Downloading attachment failed: HTTP code > 500 or '
            'weak connection. Url: ' + self.url
        )


class Photo(FileAttachment):
    tp = 'photo'
    pf_dir = 'photos'

    def __init__(self, json):
        self.url = max(json['sizes'], key=itemgetter('width', 'height'))['url']
        self.filename = f"{json['owner_id']}_{json['id']}.jpg"


class Video(Attachment):
    tp = 'video'

    def __init__(self, json):
        self.filename = sanitize_filename(
            f"{json['title']}_{json['owner_id']}_{json['id']}.mp4",
            replacement_text='_'
        )


class Audio(Attachment):
    tp = 'audio'

    def __init__(self, json):
        self.title = f"{json['artist']} - {json['title']}"


class Document(FileAttachment):
    tp = 'doc'
    pf_dir = 'documents'

    def __init__(self, json):
        self.url = json['url']
        self.filename = sanitize_filename(
            f"{json['title']}_{json['owner_id']}_{json['id']}.{json['ext']}",
            replacement_text='_'
        )


class Link(Attachment):
    tp = 'link'

    def __init__(self, json):
        self.title = json['title']
        self.url = json['url']


class Wall(Attachment):
    tp = 'wall'

    def __init__(self, json):
        self.url = f"vk.com/wall{json['from_id']}_{json['id']}"


class WallReply(Attachment):
    tp = 'wall_reply'

    def __init__(self, json):
        self.url = f"vk.com/wall{json['owner_id']}_{json['post_id']}?reply={json['id']}"


class Sticker(FileAttachment):
    tp = 'sticker'
    pf_dir = 'stickers'

    def __init__(self, json):
        self.url = max(json['images'], key=itemgetter('width', 'height'))['url']
        self.filename = f"{json['product_id']}_{json['sticker_id']}.png"


class Gift(FileAttachment):
    tp = 'gift'
    pf_dir = 'gifts'

    def __init__(self, json):
        self.url = json['thumb_256']
        self.filename = f"{json['id']}.jpg"


class AudioMessage(FileAttachment):
    tp = 'audio_message'
    pf_dir = 'audio_messages'

    def __init__(self, json):
        self.url = json['link_mp3']
        self.filename = f"{json['owner_id']}_{json['id']}.mp3"


class Graffiti(FileAttachment):
    tp = 'graffiti'
    pf_dir = 'graffiti'

    def __init__(self, json):
        self.url = json['url']
        self.filename = f"{json['owner_id']}_{json['id']}.png"


class Call(Attachment):
    tp = 'call'

    def __init__(self, json):
        self.initiator_id = json['initiator_id']

        self.call_type = ''

        if json['receiver_id'] > 2000000000:
            self.call_type += 'групповой '

        if json['video']:
            self.call_type += 'видео'

        self.state = json['state']

        self.duration = '{M}:{S}'.format(
            M=json['duration'] // 60,
            S=str(json['duration'] % 60).zfill(2)
        )

    def get_state(self, owner_id):
        if self.state == 'reached':
            return 'завершен'

        if owner_id == self.initiator_id:
            if self.state == 'canceled_by_initiator':
                return 'отменен'

            if self.state == 'canceled_by_receiver':
                return 'отклонен'

        else:
            if self.state == 'canceled_by_initiator':
                return 'пропущен'

            if self.state == 'canceled_by_receiver':
                return 'отменен'


class Poll(Attachment):
    tp = 'poll'

    def __init__(self, json):
        self.title = json['question']
        self.mult = 'несколько вариантов' if json['multiple'] else 'один вариант'
        self.anon = 'анонимный' if json['anonymous'] else 'публичный'
        self.all_votes = json['votes']
        self.answers = json['answers']
        self.answer_ids = set(json['answer_ids'])


# Словарь формата тип-вложение для более быстрого создания вложений
_attachments = {
    atch.tp: atch
    for atch in Attachment.__subclasses__() + FileAttachment.__subclasses__()
}


def gen_attachment(json):
    """
    Генерирует вложение по его JSON

    Args:
        json (dict): Объект вложения, полученный ранее благодаря VK API

    Returns:
        Attachment: Сгенерированный объект вложения
    """
    return _attachments.get(json['type'], Attachment)(json[json['type']])
